(** Integer numbers with rounding and units.

   E.g., 134 entails 13- entails 1-- entails 0---.
   and 1-k = 1-----, idem for M, G, T, P, E

*)

open Logic

module type PARAM =
  sig
    val toks_unit : Syntax.t_list
  end

module Make (Param : PARAM) =
  struct
    include Default

    let props () =
      {no_props "Intpow" with
       df = isok;
       st = isok;
       st' = isok;
       sg' = isok;
       po_entails = isok;
       cs_entails = isok;
       cp_entails = isok;
       cp'_entails = isok;
       cp_top = isok;
       cs_bot = isok;
       cs_conj = isok;
       cp_conj = isok;
       cs_disj = isok;
       cp_disj = isok;
       total_le = isok;
       po_le_min = isok;
       cs_le_min = isok;
       cp_le_min = isok;
       po_le_max = isok;
       cs_le_max = isok;
       cp_le_max = isok;
       cs_le_overlap = isok;
       cp_le_overlap = isok;
       reduced_right = isok;
       reduced_bot = isok;
       reduced_bot' = isok;
       reduced_top = isok;
     }

    type t = int * int  (* (n,p) = [n*10^p, (n+1)*10^p[, interval *)

    let rec power10 : int -> int =
      function
      | 0 -> 1
      | p -> 10 * power10 (p-1)

    let get_min : t -> int = fun (n,p) -> n * power10 p

    let get_max : t -> int = fun (n,p) -> (n+1) * power10 p

    let isvalue (n,p) = p = 0

    let compare x1 x2 =
      if get_max x1 <= get_min x2 then -1
      else if get_min x1 >= get_max x2 then 1
      else 0

    let entails x1 x2 = get_min x1 >= get_min x2 && get_max x1 <= get_max x2

    let le_min x1 x2 = get_min x1 <= get_min x2

    let le_max x1 x2 = get_max x1 <= get_max x2

    let le_overlap x y = get_min x <= get_max y

    let conj x1 x2 =
      if entails x1 x2 then x1
      else if entails x2 x1 then x2
      else raise Not_found

    let add x1 x2 = conj x1 x2

    let sub x1 x2 = if entails x1 x2 then raise Not_found else x1

    let rec features = function
      | (0,p) -> [(true,(0,p))]
      |	(n,p) -> 
	  if n < 10
	  then (true,(n,p)) :: features (n/10,p+1)
	  else features (n/10,p+1)

    let gen x1 x2 xs = []

    open Token

    let rec parse_zeros = parser
      | [<'Minus; p = parse_zeros>] -> p+1
      | [<'Ident "k">] -> 3
      | [<'Term "M">] -> 6
      | [<'Term "G">] -> 9
      | [<'Term "T">] -> 12
      | [<'Term "P">] -> 15
      | [<'Term "E">] -> 18
      | [<>] -> 0

    let parse = parser
      |	[<n = Syntax.parse_int; p = parse_zeros; _ = Syntax.parse_tokens Param.toks_unit>] -> (n,p)

    let rec print_zeros = function
      | 0 -> []
      | 3 -> [Ident "k"]
      | 6 -> [Term "M"]
      | 9 -> [Term "G"]
      | 12 -> [Term "T"]
      | 15 -> [Term "P"]
      | 18 -> [Term "E"]
      | p -> Minus :: print_zeros (p-1)

    let rec print = function
      |	(n,p) -> Syntax.toks_of_int n @ print_zeros p @ Param.toks_unit

    let simpl (n,p) =
      let n' = n / 10 in
      if n' = 0
      then [<>]
      else [<'(n',p+1)>]

    module Diff =
      struct
        type t = int * int
        let compare = compare
        let top = top
        let bot = bot
        let entails = entails
        let conj = conj
        let add = add
        let sub = sub
        let features = features
        let gen = gen
        let parse = parse
        let print = print
        let simpl = simpl
      end

    let diff x1 x2 = (get_max x1 - get_min x2, 0)

  end 
